#lang racket/base

; Evented userland for os.rkt

(require racket/set)
(require racket/match)
(require racket/list)
(require "os.rkt")

(provide (struct-out subscribe)
	 (struct-out unsubscribe)
	 (struct-out send-message)
	 (struct-out send-meta-message)
	 (struct-out spawn)

	 (struct-out on-message)
	 (struct-out on-meta-message)
	 message-handlers
	 meta-message-handlers
	 ground-message-handler

	 (except-out (struct-out transition) transition)
	 (rename-out [make-transition transition])

	 ground-vm ;; reprovided from os.rkt for convenience
	 os-big-bang)

;; A SID is an Any, a world-specific identifier for subscriptions.

;; An Action is one of
;; -- (subscribe SID EventDescription), to add a subscription
;; -- (unsubscribe SID), to remove ALL previously-added subscriptions with this SID
;; -- (send-message Message), to emit a message into the local medium
;; -- (send-meta-message MetaMessage), to emit a message into the containing medium
;; -- (spawn BootK), to start a new sibling in the raw os.rkt eventing model
(struct subscribe (sid event-description) #:transparent)
(struct unsubscribe (sid) #:transparent)
(struct send-message (body) #:transparent)
(struct send-meta-message (body) #:transparent)
(struct spawn (thunk) #:transparent)

;; An EventDescription is one of
;; -- (on-message MessagePattern (Message WorldState -> Transition)), conditionally
;;    invoked when a message arrives
;; -- (on-meta-message MetaPattern (MetaMessage WorldState -> Transition)), conditionally
;;    invoked when a metamessage arrives
;; If multiple EventDescriptions match a given event, one is
;; nondeterministically selected.
(struct on-message (pattern handler) #:transparent)
(struct on-meta-message (pattern handler) #:transparent)

;; A Transition is either
;; -- a (transition WorldState ListOf<Action>) or
;; -- a WorldState
(struct transition (state actions) #:transparent)
(define (make-transition state . actions) (transition state (flatten actions)))

;; A World is a (world WorldState Map<SID,Set<EventDescription>>), a
;; representation of a suspended world and its active subscriptions.
(struct world (state subscriptions) #:transparent)

(define-syntax message-handlers*
  (syntax-rules ()
    ((_ action-constructor old-state-pattern [pattern body ...] ...)
     (action-constructor (match-lambda [pattern #t] ... [_ #f])
			 (lambda (message old-state)
			   (match-define old-state-pattern old-state)
			   (match message
			     [pattern body ...] ...))))))

(define-syntax message-handlers
  (syntax-rules ()
    ((_ old-state-pattern [pattern body ...] ...)
     (message-handlers* on-message old-state-pattern [pattern body ...] ...))))

(define-syntax meta-message-handlers
  (syntax-rules ()
    ((_ old-state-pattern [pattern body ...] ...)
     (message-handlers* on-meta-message old-state-pattern [pattern body ...] ...))))

(define-syntax ground-message-handler
  (syntax-rules (=>)
    ((_ old-state-pattern [(tag-expr evt-expr => pattern) body ...])
     (on-meta-message (ground-event-pattern tag-expr evt-expr)
		      (lambda (meta-message old-state)
			(match-define old-state-pattern old-state)
			(match meta-message
			  [(ground-event-value _ pattern) body ...]))))))

(define (world->os-suspension w)
  (suspension w
	      #f
	      (for*/list ([(sid vs) (world-subscriptions w)]
			  [v vs]
			  #:when (on-message? v))
		(match-define (on-message pattern handler) v)
		(message-handler pattern (wrap-handler handler)))
	      (for*/list ([(sid vs) (world-subscriptions w)]
			  [v vs]
			  #:when (on-meta-message? v))
		(match-define (on-meta-message pattern handler) v)
		(message-handler pattern (wrap-handler handler)))))

(define (((wrap-handler handler) message) w)
  (maybe-transition->os-transition w (handler message (world-state w))))

(define (maybe-transition->os-transition w t)
  (if (transition? t)
      (transition->os-transition w t)
      (transition->os-transition w (transition t '()))))

(define (transition->os-transition w t)
  (match-define (transition state actions) t)
  (kernel-mode-transition (world->os-suspension (update-world w state actions))
			  (for/list [(a actions) #:when (send-message? a)]
			    (send-message-body a))
			  (for/list [(a actions) #:when (send-meta-message? a)]
			    (send-meta-message-body a))
			  (for/list [(a actions) #:when (spawn? a)]
			    (spawn-thunk a))))

(define (update-world w new-state actions)
  (world new-state
	 (foldl (lambda (action old-map)
		  (match action
		    [(subscribe sid desc)
		     (hash-update old-map sid
				  (lambda (s) (set-add s desc))
				  (lambda () (set desc)))]
		    [(unsubscribe sid)
		     (hash-remove old-map sid)]
		    [_ old-map]))
		(world-subscriptions w)
		actions)))

(define (os-big-bang initial-state . initial-actions)
  (lambda () 
    (transition->os-transition (world (void) (hash))
			       (transition initial-state (flatten initial-actions)))))
